/*
 Copyright Polyvi Inc. (http://polyvi.github.io/openxface)

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

var path = require('path'),
    shelljs = require('shelljs'),
    fs = require('fs-extra'),
    coho = require('../coho.js'),
    optimist = require('optimist'),
    reposet = require('./reposet.js');

function installXmen() {
    console.log('start installing xmen cli and its dependencies!');

    var origPath = process.cwd();

    // update lastest code for tooling repos
    if (0 !== shelljs.exec('xsrc repo-update -r tooling').code) {
        console.error('Failed to update tooling repos');
        return;
    }

    var toolingRepos = coho.getToolingRepos();
    toolingRepos.forEach(function(tool) {
        var repoPath = path.join(origPath, tool.repoName);
        if (tool.repoName === 'xface-lib') {
            repoPath = path.join(origPath, tool.repoName, 'xface-lib');
        }

        shelljs.cd(repoPath);
        shelljs.exec('npm install');
    });

    var xfacelibModulePath = path.join(origPath, 'xface-lib/xface-lib/node_modules');

    // create symbolic links
    var stats = null;
    shelljs.cd(xfacelibModulePath);
    if (fs.existsSync('xcode') &&
        ((stats = fs.lstatSync('xcode')) && !stats.isSymbolicLink())) {
        console.log('creating symbolic link to xcode for xface-lib');
        shelljs.rm('-rf', 'xcode');
        fs.symlinkSync('../../../node-xcode', 'xcode', 'junction');
    }

    ['xplugin', 'xmen-cli'].forEach(function(tool) {
        var dir = path.join(origPath, tool, 'node_modules');
        shelljs.cd(dir);

        if (fs.existsSync('xface-lib') &&
            ((stats = fs.lstatSync('xface-lib')) && !stats.isSymbolicLink())) {
            console.log('creating symbolic link to xface-lib for %s', tool);
            shelljs.rm('-rf', 'xface-lib');
            fs.symlinkSync('../../xface-lib/xface-lib', 'xface-lib', 'junction');
        }
    });

    // npm link on xmen
    shelljs.cd(path.join(origPath, 'xmen-cli'));
    shelljs.exec('npm link');
}

function installGitHook(whichHook, repos) {
    // in case git-hooks repo missed
    coho.cloneRepos([
        {repoName: 'git-hooks'}
    ], true);

    var reposetPath = reposet.root();
    var hookPath = path.join(reposetPath, 'git-hooks', whichHook);
    if (!fs.existsSync(hookPath)) {
        console.error('No such hook!');
        process.exit(1);
    }

    repos = coho.computeReposFromFlag(repos);

    shelljs.cd(hookPath);
    repos.forEach(function (repo) {
        shelljs.exec('python hooked.py --git-root=' + path.join(reposetPath, repo.repoName));
    });
}

function installHookCommand() {
    var opt = optimist.options('k', {
        alias: 'hook',
        desc: 'Which hook to be installed.'
    });
    opt = coho.registerRepoFlag(coho.registerHelpFlag(opt));
    var argv = opt
        .usage('Installs git hooks to specified repos, e.g., \n\n' +
            '    xsrc hook-install -k pre-commit -r plugins')
        .argv;

    if (argv.h) {
        optimist.showHelp();
        process.exit(1);
    }

    installGitHook(argv.k, argv.r);
}

module.exports = {
    commands: [{
        name: 'xmen-install',
        desc: 'Install xmen from reposet',
        entryPoint: installXmen
    },
    {
        name: 'hook-install',
        desc: 'Install git hooks from git-hooks repo',
        entryPoint: installHookCommand
    }]
}
