/*
 Copyright Polyvi Inc. (http://polyvi.github.io/openxface)

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

var fs = require('fs'),
    coho = require('../coho.js'),
    optimist = require('optimist'),
    xml_helpers = require('./xml-helpers'),
    semver = require('semver'),
    path = require('path'),
    Q = require('q'),
    prompt = require('prompt'),
    shell = require('shelljs');

function releaseCommand() {
    var opt = coho.registerHelpFlag(optimist);
    var argv = opt
        .usage('Do release operation, e.g., \n\n' +
            '    xsrc release -r xface-extension-xxx [--version=x.x.x]\n' +
            'Make sure you carefully read the link\n' +
            'http://twiki.polyvi.com/twiki5/bin/view/Dev/ProjApolloV32PluginRelease' +
            'and http://twiki.polyvi.com/twiki5/bin/view/Dev/ProjApolloV32CLIRelease')
        .argv;

    if (!argv.r || argv.h || (argv.version && typeof argv.version != 'string')) {
        optimist.showHelp();
        process.exit(1);
    }
    // check version parameter
    if(argv.version && !/^\d+\.\d+\.(\d+|\d+-(rc)?[\d.]*\d)$/.test(argv.version)) {
        coho.fatal('Versions must be in the form #.#.#[-rc#]');
    }

    var repoNames = [];
    var repos = coho.computeReposFromFlag(argv.r);
    // First - perform precondition checks.
    coho.updateRepos(repos, [], true);

    // js, default-app will release before platform
    // for cli, the order is xplugin -> xface-cli -> xmen-cli
    moveRepoToFirst(repos, 'xface-cli');
    moveRepoToFirst(repos, 'xplugin');
    moveRepoToFirst(repos, 'default-app');
    moveRepoToFirst(repos, 'js');

    coho.forEachRepoAsync(repos, function(repo) {
        var repoName = repo.repoName;
        console.log('\nStart to release repo "' + repoName + '".....');
        repoNames.push(repoName);
        if(/^xface-extension-|xface-extra-|cordova-plugin-/i.test(repoName)) {
            return releasePlugin(repo, argv.version);
        } else if(/^xface-(android|ios|wp8|js|test-template|default-app|test-apps)$/i.test(repoName)) {
            return releasePlatform(repo, argv.version);
        } else if(/xmen-cli|xface-cli|xplugin|xsrc/.test(repoName)) {
            return releaseCliTool(repo, argv.version);
        } else {
            console.warn('Release process for repo "' + repoName + '" is not supported, just ingnored.....');
            repoNames.pop();
            return Q();
        }
    })
    .done(function() {
        console.warn('\n\nNow you need check repos ' + require('util').format(repoNames)
            + ', then push commits and tag to remote repository.');
    }, function(err) {
        throw err;
    });
}

function calculatePlatformVersion(repoName) {
    var version = getLatestTag();
    var index = version.indexOf('rc');
    index == -1 && (index = version.indexOf('dev'));
    if(version.indexOf('-') != -1) {
        version = semver.inc(version, 'prerelease');
    } else if(index != -1) {
        version = version.substr(0, index) + '-' + version.substr(index);
        version = semver.inc(version, 'prerelease');
    } else {
        version = semver.inc(version, 'patch');
    }
    // correct the new version, for semver, 3.2.0-rc1 will become 3.2.0-rc1.0 after `inc` operation
    // we need change 3.2.0-rc1.0 to 3.2.0-rc2
    var matchData;
    if(!!(matchData = version.match(/^(.*-[a-z]+)(\d+)\.0$/i))) {
        version = matchData[1] + (parseInt(matchData[2]) + 1);
    }
    return promptVersion('platform', '[Input the platform release version]', repoName, version);
}

var savedPlatformReleaseVersion;
var dependentRepoUpdated;

function releasePlatform(repo, version) {
    var repoName = repo.repoName;

    if(savedPlatformReleaseVersion && !version) {
        version = savedPlatformReleaseVersion;
    }
    var q = version ? Q(version) : calculatePlatformVersion(repoName);
    return q.then(function(version) {
        version = coho.validateVersionString(version);
        return checkVersionTagExisted(repoName, version);
    })
    .then(function(version) {
        var branchName = coho.getVersionBranchName(version);
        !savedPlatformReleaseVersion && (savedPlatformReleaseVersion = version);

        coho.stashAndPop(repo, function() {
            // git fetch + update master
            coho.updateRepos([repo], ['master'], false);

            // 1. update default app and image resouce on branch 'master'
            if(/^xface-(android|ios|wp8)$/i.test(repoName)) {
                // maybe repo 'js' and 'default-app' are released on other computer
                // so we need update them first before release xface engine
                if(!dependentRepoUpdated) {
                    var dependentRepos = coho.computeReposFromFlag(['js', 'default-app']);
                    coho.updateRepos(dependentRepos, ['master', branchName]);
                    dependentRepoUpdated = true;
                }
                var platform = repoName.slice('xface-'.length);
                var destAppPaths = {
                    'android' : path.join('bin', 'templates', 'project', 'assets', 'xface3'),
                    'ios' : path.join('bin', 'templates', 'project', 'xface3'),
                    'wp8' : path.join('bin', 'template', 'xface3')
                }
                var defaultAppRepoPath = path.join('..', 'xface-default-app');
                if(!fs.existsSync(defaultAppRepoPath)) {
                    console.warn('Can\'t find repository at path "' + path.resolve(defaultAppRepoPath) + '", continue...');
                } else {
                    shell.cp('-rf', path.join(defaultAppRepoPath, 'www', 'helloxface'), destAppPaths[platform]);
                    copyResources(platform, defaultAppRepoPath);
                }
                coho.execHelper('git add .');
                // if changed something, commit first
                if(coho.pendingChangesExist()) {
                    coho.execHelper('git commit -am "Update default app and resources."');
                }
            }

            // 2. Either create or pull down the release branch.
            if (coho.remoteBranchExists(repo, branchName)) {
                console.log('Remote branch already exists for repo: ' + repoName);
                // Check out and rebase.
                coho.updateRepos([repo], [branchName], true);
                coho.checkoutBranch(branchName);
            } else if (coho.localBranchExists(branchName)) {
                coho.execHelper('git checkout ' + branchName);
            } else {
                coho.checkoutBranch('master');
                coho.execHelper('git checkout -b ' + branchName);
            }

            // 3. merge release branch with master
            coho.execHelper('git merge -X theirs master');

            // 4. update js for release branch, write release notes, then update version
            coho.updateJsSnapshot(repo, version);

            // generate release note only for final release
            if (!/rc/.test(version)) {
                var releaseNotesPath = 'RELEASENOTEX.md';
                writeReleaseHeader(releaseNotesPath, version);
                generateReleaseNotes(releaseNotesPath, getLatestTag(), branchName);
            }
            coho.execHelper('git add .');
            console.log(repoName + ': ' + 'Setting VERSION to "' + version + '" on branch + "' + branchName + '".');
            coho.updateRepoVersion(repo, version);

            // 5. tag
            console.log('Tagging %s to %s', repoName, version);
            coho.execHelper('git tag ' + version + ' --force');

            // 6. update version and js for master branch
            coho.checkoutBranch('master');
            var devVersion = coho.createPlatformDevVersion(version);
            console.log(repoName + ': ' + 'Setting VERSION to "' + devVersion + '" on branch + "master".');
            coho.execHelper('git add .');
            coho.updateRepoVersion(repo, devVersion);
            coho.updateJsSnapshot(repo, devVersion);
        });
    });
}

function copyResources(platform, defaultAppRepoPath) {
    var resources = {
        'android' : [
            {
                srcDir : path.join(defaultAppRepoPath, 'res', 'screen', 'android'),
                destDir : path.join('bin', 'templates', 'project', 'res'),
                map : {
                    'screen-hdpi-landscape' : path.join('drawable-land-hdpi', 'screen'),
                    'screen-ldpi-landscape' : path.join('drawable-land-ldpi', 'screen'),
                    'screen-mdpi-landscape' : path.join('drawable-land-mdpi', 'screen'),
                    'screen-xhdpi-landscape' : path.join('drawable-land-xhdpi', 'screen'),
                    'screen-hdpi-portrait' : path.join('drawable-port-hdpi', 'screen'),
                    'screen-ldpi-portrait' : path.join('drawable-port-ldpi', 'screen'),
                    'screen-mdpi-portrait' : path.join('drawable-port-mdpi', 'screen'),
                    'screen-xhdpi-portrait' : path.join('drawable-port-xhdpi', 'screen')
                }
            },
            {
                srcDir : path.join(defaultAppRepoPath, 'res', 'icon', 'android'),
                destDir : path.join('bin', 'templates', 'project', 'res'),
                map : {
                    'icon-36-ldpi' : path.join('drawable-ldpi', 'icon'),
                    'icon-48-mdpi' : path.join('drawable-mdpi', 'icon'),
                    'icon-72-hdpi' : path.join('drawable-hdpi', 'icon'),
                    'icon-72-hdpi' : path.join('drawable', 'icon'),
                    'icon-96-xhdpi' : path.join('drawable-xhdpi', 'icon')
                }
            }
        ],
        'ios' : [
            {
                srcDir : path.join(defaultAppRepoPath, 'res', 'screen', 'ios'),
                destDir : path.join('bin', 'templates', 'project', '__PROJECT_NAME__', 'Resources', 'splash'),
                map : {
                    'screen-ipad-landscape' : 'Default-Landscape~ipad',
                    'screen-ipad-landscape-2x' : 'Default-Landscape@2x~ipad',
                    'screen-ipad-portrait' : 'Default-Portrait~ipad',
                    'screen-ipad-portrait-2x' : 'Default-Portrait@2x~ipad',
                    'screen-iphone-portrait' : 'Default~iphone',
                    'screen-iphone-portrait-2x' : 'Default@2x~iphone',
                    'screen-iphone-portrait-568h-2x' : 'Default-568h@2x~iphone'
                }
            },
            {
                srcDir : path.join(defaultAppRepoPath, 'res', 'icon', 'ios'),
                destDir : path.join('bin', 'templates', 'project', '__PROJECT_NAME__', 'Resources', 'icons'),
                map : {
                    'icon-57' : 'icon',
                    'icon-57-2x' : 'icon@2x',
                    'icon-60' : 'icon-60',
                    'icon-60-2x' : 'icon-60@2x',
                    'icon-72' : 'icon-72',
                    'icon-72-2x' : 'icon-72@2x',
                    'icon-76' : 'icon-76',
                    'icon-76-2x' : 'icon-76@2x'
                }
            }
        ],
        'wp8' : [
            {
                srcDir : path.join(defaultAppRepoPath, 'res', 'screen', 'windows-phone'),
                destDir : path.join('bin', 'template')
            },
            {
                srcDir : path.join(defaultAppRepoPath, 'res', 'icon', 'windows-phone'),
                destDir : path.join('bin', 'template', 'Assets')
            }
        ]
    };

    var operationSet = resources[platform];
    operationSet.forEach(function(op) {
        if(op.map) {
            Object.keys(op.map).forEach(function(from) {
                shell.cp('-f', path.join(op.srcDir, from + '.png'), path.join(op.destDir, op.map[from] + '.png'));
            });
        } else {
            shell.cp('rf', path.join(op.srcDir, '*'), op.destDir);
        }
    });
}

// refer to http://twiki.polyvi.com/twiki5/bin/view/Dev/ProjApolloV32PluginRelease
function releasePlugin(repo, version) {
    var repoName = repo.repoName;
    // workaround: don't release premature cordova-lab-plugins
    if (/keyboard/i.test(repoName)) {
        return;
    }

    var releaseNotesPath = /xface/i.test(repoName) ? 'RELEASENOTES.md' : 'RELEASENOTEX.md';
    var pluginxml = './plugin.xml';

    // make sure we're code fresh and on dev branch
    coho.updateRepos([repo], ['master', 'dev']);
    coho.execHelper('git checkout dev');

    // 1. update releasenotes.md & version
    var xml = xml_helpers.parseElementtreeSync(pluginxml);
    var q = Q(version);
    !version && (q = generatePluginReleaseVersion(repoName, xml.getroot().attrib.version));
    return q.then(function(version) {
        return checkVersionTagExisted(repoName, version, 'r' + version);
    })
    .then(function(v) {
        version = v;
        updatePluginVersion(pluginxml, version);

        writeReleaseHeader(releaseNotesPath, version);
        // write commit messages
        generateReleaseNotes(releaseNotesPath, 'master', 'dev');

        // commit
        coho.execHelper('git commit -am "Updated version and RELEASENOTES.md for release '
            + version + '"');

        // 2. merge changes into master & tag
        mergeAndTagPlugin(repo, version);

        // 3. Update dev branch's version
        coho.execHelper('git checkout dev');
        // increment version by release type 'patch', 1.0.0 -> 1.0.1, plus '-dev'
        var new_version = semver.inc(version, 'patch') + '-dev';
        updatePluginVersion(pluginxml, new_version);
        coho.execHelper('git commit -am "Incremented plugin version on dev branch to '
            + new_version + '"');
    });
}

function moveRepoToFirst(repos, id) {
    var xfaceJsRepo = coho.getRepoById(id);
    var repoIndex = repos.indexOf(xfaceJsRepo);
    if (repoIndex != -1) {
        repos.splice(repoIndex, 1);
        repos.unshift(xfaceJsRepo);
    }
}

function getLatestTag(repoGitPath) {
    var gitDirOption = repoGitPath ? '--git-dir=' + repoGitPath : '';
    var sha1Id = coho.execHelper('git ' + gitDirOption + ' rev-list --tags --max-count=1', true);
    return coho.execHelper('git ' + gitDirOption + ' describe --tags ' + sha1Id, true);
}

// if release version is not provided, we calculate it and then prompt user to change it
function generatePluginReleaseVersion(repoName, devVersion) {
    var releaseVersion;
    var index = devVersion.indexOf('-dev');
    // remove '-dev' if available
    if (index > -1) {
        releaseVersion = devVersion.slice(0, index);
    } else {
        releaseVersion = semver.inc(devVersion, 'patch');
    }
    return promptVersion('plugin', '[Input the plugin release version]', repoName, releaseVersion);
}

function updatePluginVersion(pluginxml, newVersion) {
    var content = fs.readFileSync(pluginxml, 'utf8');
    fs.writeFileSync(pluginxml, content.replace(/(<plugin[\s\S]*?version\s*=\s*").*?("\s*>)/,
        '$1' + newVersion + '$2'), 'utf8');
}

function writeReleaseHeader(file, version) {
    fs.appendFileSync(file, '\n\n## ' + version + ' (' + new Date().toDateString() + ')', 'utf-8');
}

function generateReleaseNotes(file, since, until) {
    var cmd = 'git log --pretty=format:" * % s" --topo-order --no-merges ';
    if(since) {
        cmd += (since + '..');
    }
    cmd += until;
    var output = coho.execHelper(cmd, true);
    // remove trello action words and issue identifier #
    if(!output) {
        console.warn('The repo "' + path.resolve(process.cwd()) + '" was not changed since last version ' + since);
    } else {
        fs.appendFileSync(file, '\n\n\n ' + output.replace(/(\[(doing|done|close)-\d+\]\s*)/ig, '').
            replace(/#\s*(\d+)/g, 'issue $1').trim() + '\n', 'utf-8');
    }
}

function mergeAndTagPlugin(repo, version) {
    coho.execHelper('git checkout master');
    coho.execHelper('git merge dev --ff-only');
    console.log('Tagging %s to %s', repo.repoName, version);
    coho.execHelper('git tag r' + version + ' --force');
}

function releaseCliTool(repo, version) {
    var repoName = repo.repoName;
    coho.updateRepos([repo], ['master']);
    coho.execHelper('git checkout master');

    var packageJson = JSON.parse(fs.readFileSync('package.json', 'utf-8')),
        oldVersion = packageJson.version;
    !version && (version = semver.inc(oldVersion, 'patch'));
    return promptVersion('cli', '[Input the cli release version]', repoName, version)
    .then(function(v) {
        version = v;
        return checkVersionTagExisted(repoName, version)
    })
    .then(function() {
        // write to RELEASENOTEX.md
        writeReleaseHeader('./RELEASENOTEX.md', version);
        if(coho.tagExists(oldVersion)) {
            generateReleaseNotes('./RELEASENOTEX.md', oldVersion, 'master');
        } else {
            generateReleaseNotes('./RELEASENOTEX.md', null, 'master');
        }
    })
    .then(function() {
        // update version of package.json, include dependencies
        packageJson.version = version;
        var dependencies = packageJson['dependencies'];
        var dependencyMap = {'xface' : 'xface-cli', 'xplugin' : 'xplugin'};

        return Object.keys(dependencyMap).reduce(function(soFar, name) {
            return soFar.then(function() {
                if(!dependencies[name]) return;

                var jsonFile = path.join('..', dependencyMap[name], 'package.json');
                if(!fs.existsSync(jsonFile)) {
                    console.log('Can\'t find repo "' + dependencyMap[name] + '" in reposet.');
                    return;
                }
                var dependencyVersion = JSON.parse(fs.readFileSync(jsonFile, 'utf-8'))['version'];
                return promptVersion('cli', '[Input the dependent npm module version]', name, dependencyVersion)
                .then(function(v) {
                    dependencies[name] = v;
                });
            });
        }, Q())
        .then(function() {
            fs.writeFileSync('package.json', JSON.stringify(packageJson, null, 2), 'utf-8');
        });
    })
    .then(function() {
        // modify platforms.js if xface-cli
        if(repoName == 'xface-cli') {
            // fetch the latest tag name of platform
            var platformPath = path.join('..', 'xface-android');
            !fs.existsSync(platformPath) && (platformPath = path.join('..', 'xface-ios'));
            var latestPlatformVersion;
            if(!fs.existsSync(platformPath)) {
                console.warn('Can\'t find platform repo "' + path.resolve(platformPath) + '".');
            } else {
                var repoGitPath = path.join(platformPath, '.git');
                latestPlatformVersion = getLatestTag(repoGitPath);
            }

            return promptVersion('platform', '[Input the dependent platform version]', 'android/ios/wp8', latestPlatformVersion)
            .then(function(v) {
                latestPlatformVersion = v;
                var content = fs.readFileSync('platforms.js', 'utf-8');
                var platforms = ['android', 'ios', 'wp8', 'www', 'test-template'];
                platforms.forEach(function(p) {
                    content = modifyPlatformVersion(content, p, latestPlatformVersion);
                });
                fs.writeFileSync('platforms.js', content, 'utf-8');
            });
        }
    })
    .then(function() {
        // commit
        coho.execHelper('git add .');
        coho.execHelper('git commit -am "Updated version and RELEASENOTES.md for release '
            + version + '"');
        console.log('Tagging %s to %s', repoName, version);
        coho.execHelper('git tag ' + version + ' --force');
    });
}

function checkVersionTagExisted(repoName, version, tag) {
    !tag && (tag = version);
    if(coho.tagExists(tag)) {
        return promptOverrideTag(repoName, version)
        .then(function(force) {
            if(force) {
                return Q(version);
            } else {
                // quit
                coho.fatal('Version "' + version + '" is existed in current repository, quit the release process.');
            }
        });
    }
    return Q(version);
}

// Prompt user to override the old tag existed
function promptOverrideTag(repoName, tagName) {
    prompt.start();
    prompt.message = 'Tag "' + tagName + '" is already existed in the current repo "' + repoName + '", override it? ';
    return Q.nfcall(prompt.get, [{
        name: 'force',
        description: '[yes or no]',
        default: 'yes',
        type: 'string',
        required: true,
        pattern: /^(y|yes|n|no)$/i
    }])
    .then(function(result) {
        return /y|yes/i.test(result.force);
    });
}

// Prompt user to change version value
function promptVersion(type, message, description, defaultValue) {
    prompt.start();
    prompt.message = message;
    return Q.nfcall(prompt.get, [{
        name: 'version',
        description: description,
        default: defaultValue,
        type: 'string',
        required: true,
        conform: function(v) {
            if(type === 'platform' || type === 'plugin') {
                return /^\d+\.\d+\.\d+-?/.test(v);
            } else if(type === 'cli') {
                return /^\d+\.\d+\.(\d+|x)-?/.test(v);
            } else {
                return true;
            }
        }
    }])
    .then(function(result) {
        return result.version;
    });
}

function modifyPlatformVersion(content, plat, version) {
    var expr = RegExp('(\'' + plat + '\'[\\s\\S]*?version\\s*:\\s*\').*?(\')');
    return content.replace(expr, '$1' + version + '$2');
}

module.exports = {
    commands: [
        {
            name: 'release',
            desc: 'Do release on repos',
            entryPoint: releaseCommand
        }
    ]
};
