/*
 Copyright Polyvi Inc. (http://polyvi.github.io/openxface)

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

var optimist = require('optimist'),
    path = require('path'),
    fs = require('fs-extra'),
    coho = require('../coho.js'),
    shell = require('shelljs'),
    util = require('./utils.js');

function repoCommand() {
    var opt = optimist.options('i', {
        alias: 'init',
        desc: 'Initialize repo with template.'
    }).options('c', {
        alias: 'checkout',
        desc: 'Check out repos to specific versions.'
    });

    var argv = optimist.argv;
    if (argv.i) {
        return initWithTemplate();
    }

    if (argv.c) {
        return checkout();
    }

    // set command major help
    opt.usage('Repo related commands, e.g., \n\n' +
            '    xsrc repo -i -n xface-xxx -l gplv3\n' +
            '    xsrc repo -c -r xface-android@3.2.x -r xface-extension-device');

    optimist.showHelp();
    process.exit(1);
}

function initWithTemplate() {
    var opt = optimist
        .options('n', {
            alias: 'name',
            desc: 'What name of repo.',
            demand: 'n'
        })
        .options('l', {
            alias: 'license',
            desc: 'Which license should be applied, MUST be apache or gplv3',
            demand: 'l'
        })
        .options('dev', {
            type: 'boolean',
            desc: 'Use --no-dev to skip dev branch creation',
            default: true
        })
        .check(function (value) {
            // must be apache or gplv3
            return /^(apache|gplv3)$/i.test(value.l);
        });
    opt = coho.registerHelpFlag(opt);
    var argv = opt
        .usage('Initialize new repo with template, e.g., \n\n' +
            '    xsrc repo -i -n xface-xxx -l gplv3\n' +
            '    xsrc repo -i -n xface-xxx -l apache --no-dev\n\n' +
            'Run this command right after repo created on git server')
        .argv;

    if (argv.h) {
        optimist.showHelp();
        process.exit(1);
    }

    // clone empty git repo
    shell.exec('git clone git@gitlab.polyvi.com:xface/' + argv.n + '.git');

    // handle template
    var newRepoPath = path.join(require('./reposet.js').root(), argv.n);
    var tempaltePath = path.join(__dirname, '..', 'metadata');

    ['NOTICE', 'README.md', 'RELEASENOTES.md'].forEach(function (file) {
        fs.copySync(path.join(tempaltePath, file), path.join(newRepoPath, file));
    });

    fs.copySync(path.join(tempaltePath, 'git_ignore'), path.join(newRepoPath, '.gitignore'))

    var licenseFile;
    if (argv.l === 'apache') {
        licenseFile = 'Apachev2_LICENSE';
    } else {
        licenseFile = 'GPLv3_LICENSE';
    }
    fs.copySync(path.join(tempaltePath, licenseFile), path.join(newRepoPath, 'LICENSE'));

    var content = fs.readFileSync(path.join(tempaltePath, '.reviewboardrc'), 'utf8');
    fs.writeFileSync(path.join(newRepoPath, '.reviewboardrc'), content.replace(/xface-xxx/m, argv.n), 'utf8');

    // make first commit
    shell.cd(newRepoPath);
    coho.execHelper('git add .');
    coho.execHelper('git commit -m "Initial commit by xsrc"');

    if (argv.dev) {
        coho.execHelper('git checkout -b dev');
    }

    console.log('You still have to\n' +
        '1. notify luoq|lis|liumn to add it to reviewboard\n' +
        '2. add it to repo xface-repos\n');
}

function checkout() {
    var opt = optimist.options('r', {
        alias: 'repo',
        desc: 'Which repos to build on. Follow version with appending "@", like xxx@1.0.0.' +
            ' Multiple flags allowed. This can be repo IDs, not group ID.' +
            ' Use the list-repos command see valid values.'
    });
    opt = coho.registerHelpFlag(opt);
    var argv = opt
        .usage('Prepare reposet to be specified versions with following commands:\n' +
            '   clone missing repos\n' +
            '   update to latest\n' +
            '   checkout specific branch or commit\n' +
            '\n' +
            'Usage: $0 reposet -c -r xface-android@3.2.x -r xface-extension-device\n')
        .argv;

    if (argv.h) {
        optimist.showHelp();
        process.exit(1);
    }

    var repos = parseRepos(Array.isArray(argv.r) ? argv.r : [argv.r]);

    batchCheckout(repos, 'release', true);
}

var checkoutPolicy = {
    dev: batchCheckoutDev,
    release: batchCheckoutRelease
};

function batchCheckout(repos, policy, update) {
    // clone missing repos first
    coho.cloneRepos(repos, true);

    checkoutPolicy[policy](repos, update);
}

function isCadenceRepo(repoName) {
    var notFound = coho.getCadenceGroup().every(function(r) {
        return r.repoName !== repoName;
    });

    return !notFound;
}

function remoteBranchExists(r, name) {
    return !!coho.execHelper('git branch -r --list ' + 'origin/' + name, true);
}

function batchCheckoutDev(repos, update) {
    coho.forEachRepo(repos, function(r) {
        var branch = 'dev';

        if (isCadenceRepo(r.repoName)) {
            branch = 'master';
        }

        // some repos don't follow the rule cadence or master/dev flow,
        // only master branch, like cordova-plugin-keyboard/status
        if (remoteBranchExists(r, branch)) {
            coho.checkoutBranch(branch);
        }

        if (update) {
            coho.updateRepos([r], [branch]);
        }
    });
}

function batchCheckoutRelease(repos, update) {
    // make sure on right branch/tag/commit
    coho.forEachRepo(repos, function (repo) {
        if (update) {
            coho.execHelper('git fetch --progress');
            coho.execHelper('git fetch --tags');
        }

        if (repo.versionTag) {
            coho.execHelper('git checkout -q ' + repo.versionTag);
        } else {
            var branch = parseReleaseBranch(repo);

            // in case I'm on a detached tag, updateRepos will fail
            // so checkout to a named branch first
            coho.checkoutBranch(branch);

            if (update) {
                // make sure branch is fresh
                coho.updateRepos([repo], [branch], true);
            }
        }
    });
}

function parseReleaseBranch(repo) {
    // if no tag specified, we check out latest release branch
    // for platforms/js/default-app, it's sth like 3.2.x
    var latestReleaseBranch = 'master';
    if (isCadenceRepo(repo.repoName)) {
        var greaterVersion = 0;
        util.getRemoteBranchNames().forEach(function (b) {
            var ver = parseFloat(b);
            if (!Number.isNaN(ver)) {
                if (ver > greaterVersion) {
                    greaterVersion = ver;
                    latestReleaseBranch = b;
                }
            }
        });
    }

    return latestReleaseBranch;
}

/**
 * Read tag from input, convert to a repo object
 * @param repos array with item like ['platform', 'plugin-device@1.0.0', 'js']
 * @returns {Array} array contains converted repo objects
 */
function parseRepos(repos) {
    var result = [];

    repos.forEach(function(repo) {
        var splits = repo.split('@');
        var whichRepos = coho.computeReposFromFlag(splits[0]);
        if (splits.length === 2) {
            whichRepos.forEach(function (r) {
                r.versionTag = splits[1];
            });
        }

        result = result.concat(whichRepos);
    });

    return result;
}

module.exports = {
    commands: [{
        name: 'repo',
        desc: 'Operations with repo',
        entryPoint: repoCommand
    }],
    batchCheckout: batchCheckout
}
