/*
 Copyright Polyvi Inc. (http://polyvi.github.io/openxface)

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

var optimist = require('optimist'),
    path = require('path'),
    fs = require('fs-extra'),
    coho = require('../coho.js');

function reposetCommand() {
    var opt = optimist.options('l', {
        alias: 'list',
        desc: 'List infomation about current reposet.'
    }).options('i', {
            alias: 'init',
            desc: 'Init reposet at specified path (by -p), if no input, it\'s current directory, ' +
                're-running it will overwrite old reposet.'
        }).describe('co', 'Check out into dev or release state, use --update optionally');

    var argv = optimist.argv;
    if (argv.l) {
        return listInfo();
    }

    if (argv.i) {
        var where = process.cwd();
        if (argv.p) {
            where = path.resolve(where, argv.p);
        }
        return init(where);
    }

    if (argv.co) {
        return checkout();
    }

    // if no command matched, show help
    opt.usage('Reposet related commands, e.g., \n\n' +
            '    xsrc reposet [-l|--list]\n' +
            '    xsrc reposet -i -p /path/to/new_reposet\n' +
            '    xsrc reposet --co=dev [--update]');

    optimist.showHelp();
    process.exit(1);
}

function init(where) {
    var opt = optimist
        .options('p', {
        alias: 'path',
        desc: 'Path to where reposet get initialized. This is optional'
    });

    opt = coho.registerHelpFlag(opt);
    var argv = opt
        .usage('Initialize reposet at new path, e.g., \n\n' +
            '    xsrc reposet -i -p /path/to/new_reposet/')
        .argv;

    if (argv.h) {
        optimist.showHelp();
        process.exit(1);
    }

    if (!fs.existsSync(where)) {
        coho.fatal('%s was NOT existed!', where);
    }

    setRoot(where);
    console.log('reposet was initialized at %s', where);
}

function listInfo() {
    console.log('reposet is at %s', getRoot());
    console.warn('todo: list reposet info');
}

function getGlobalSetting() {
    var userHome = process.env[(process.platform === 'win32') ? 'USERPROFILE' : 'HOME'];
    var settingFile = path.join(userHome, '.xface', 'global_setting.json');

    if (!fs.existsSync(settingFile)) {
        fs.createFileSync(settingFile);
        fs.writeJsonSync(settingFile, {});
    }

    return settingFile;
}

function getRoot() {
    var settingFile = getGlobalSetting();
    var setting = fs.readJsonSync(settingFile);

    var root = '';
    if (setting && setting.repoSet) {
        root = setting.repoSet;
    }

    return root;
}

function setRoot(where) {
    var settingFile = getGlobalSetting();
    var setting = fs.readJsonSync(settingFile);
    setting['repoSet'] = where;
    fs.writeJsonSync(settingFile, setting);
}

function checkout() {
    var argv = optimist.boolean('update').argv;
    require('./repo.js').batchCheckout(coho.getAllRepos(), argv.co, argv.update);
}

// added *keepCwd* property, 'reposet -i' is not supposed to get executed
// under current reposet, instead it works on current directory
module.exports = {
    commands: [{
        name: 'reposet',
        desc: 'list or init reposet',
        entryPoint: reposetCommand,
        keepCwd: (optimist.argv.i) ? true : false
    }],
    root: getRoot
}
